(function() {

//// Event handling

	var Event = function(name) {
		this.name = name;
		var listeners = [];
    
		this.addListener = function(listener) {
			var i = listeners.indexOf(listener);
			if (i < 0)
				listeners.push(listener);
		};
    
		this.removeListener = function(listener) {
			if (!listener)
				listeners = [];
			else {
				var i = listeners.indexOf(listener);
				if (i < 0)
					listeners.splice(i,1);
			}
		};
    
		this.fire = function() {
			var args = Array.prototype.slice.call(arguments);
			trace("Firing event "+this.name+"("+args.join()+")  "+listeners.length+" listeners");
			for (var i=0; i<listeners.length; i++) {
				listeners[i].apply(null, args);
			}
		};
	};


//// MRAID variables

	var states = {
		loading: "loading",
		_default: "default", // ...default is reserved word
		expanded: "expanded",
		hidden: "hidden"
	};
	var state = states.loading;

	var placementType = "inline";  // ...unless overridden by setReady...
	var isViewable = false;

	var events = {
		ready: new Event("ready"),
		error: new Event("error"),
		stateChange: new Event("stateChange"),
		viewableChange: new Event("viewableChange")
	};

	var expandProperties = {
		width: -1,
		height: -1,
		useCustomClose: false
	};

//// Bridge for SDK-to-JS and JS-to-SDK functions

	var mraidBridge = window.mraidBridge = {};
	
	var pendingSdkCommands = [];
	var sdkExecutingCommand = false;
	var tracing = false;

	var trace = function(message) {
		if (!tracing) return;
		
		mraidBridge.executeInSdk("trace",encodeURIComponent(message));
	}

	var traceCall = function() {
		if (!tracing) return;
		
		var args = Array.prototype.slice.call(arguments);
		var message = "Called: "+args.shift()+"("+args.join()+")";
		mraidBridge.executeInSdk("trace",encodeURIComponent(message));
	}

	var error = function(mraidFunction, message) {
		trace("Error calling function "+mraidFunction+": "+message);
		events.error.fire(message, mraidFunction);
	};

	mraidBridge.setTracing = function(newTracing) {
		tracing = newTracing;
	};
	
	mraidBridge.setState = function(newState) {
		if (newState != state) {
			state = newState;
			events.stateChange.fire(state);
		}
	};

	mraidBridge.setIsViewable = function(newIsViewable) {
		if (newIsViewable != isViewable) {
			isViewable = newIsViewable;
			events.viewableChange.fire(isViewable);
		}
	};

	mraidBridge.useCustomClose = function(newUseCustomClose) {
		if (newUseCustomClose != expandProperties.useCustomClose) {
			expandProperties.useCustomClose = newUseCustomClose;
			mraidBridge.executeInSdk("useCustomClose",newUseCustomClose);
		}
	};

	mraidBridge.executeInSdk = function() {
		var args = Array.prototype.slice.call(arguments);
		var sdkCommand = "mraid://"+args.join("/");
		
		if (sdkExecutingCommand)
			pendingSdkCommands.push(sdkCommand);
		else {
			sdkExecutingCommand = true;
			window.location = sdkCommand;
		}
	};

	mraidBridge.sdkExecutionDone = function() {
		if (pendingSdkCommands.length == 0) {
			sdkExecutingCommand = false;
		} else {
			var sdkCommand = pendingSdkCommands.shift();
			window.location = sdkCommand;
		}
	}
	
	mraidBridge.sdkReady = function(width,height,isInterstitial,isVisible,isTracing) {
		tracing = isTracing;
	
		expandProperties.width = width;  //...the default expansion size
		expandProperties.height = height;
		expandProperties.useCustomClose = false;

		if (isInterstitial)
			placementType = "interstitial";

		events.ready.fire();
		this.setState(states._default);
		this.setIsViewable(isVisible);
	}
	
//// MRAID API functions ////

	var mraid = window.mraid = {};
	
	mraid.getVersion = function() {
		traceCall("getVersion",arguments);
		return "1.0";
	};

	mraid.getState = function() {
		traceCall("getState");
		
		return state;
	};

	mraid.getPlacementType = function() {
		traceCall("getPlacementType");
		
		return placementType;
	};

	mraid.isViewable = function() {
		traceCall("isViewable");
		
		return isViewable;
	};

	mraid.expand = function(url) {
		traceCall("expand",url);
		
		if (!url || (String(url).length==0))
			url = "x"; // Pass a single char as an indicator of no url.  Otherwise "//" is passed as part of the url and is misparsed by the SDK.
		
		if (state != states._default)
			error("expand", "An ad can only be expanded from the default state.");
		else
			mraidBridge.executeInSdk("expand",encodeURIComponent(url),
				expandProperties.width, expandProperties.height, expandProperties.useCustomClose);
	};

	mraid.open = function(url) {
		traceCall("open",url);
		
		if (!url)
			error("open", "No URL was given.");
		else
			mraidBridge.executeInSdk("open", encodeURIComponent(url));
	};

	mraid.close = function() {
		traceCall("close");
		
		if (state == states.hidden)
			error("close", "The ad is hidden.");
		else
			mraidBridge.executeInSdk("close");
	};

	mraid.getExpandProperties = function() {
		traceCall("getExpandProperties");

		return {
			width: expandProperties.width,
			height: expandProperties.height,
			useCustomClose: expandProperties.useCustomClose,
			isModal: true
		};
	};

	mraid.setExpandProperties = function(properties) {
		traceCall("setExpandProperties",properties);
		
		if (properties.hasOwnProperty("width")) expandProperties.width = properties.width;
		if (properties.hasOwnProperty("height")) expandProperties.height = properties.height;
		if (properties.hasOwnProperty("useCustomClose")) mraidBridge.useCustomClose(properties.useCustomClose);
	};

	mraid.useCustomClose = function(useCustomClose) {
		traceCall("useCustomClose",useCustomClose);
		
		mraidBridge.useCustomClose(useCustomClose);
	};

	mraid.addEventListener = function(eventType, listener) {
		traceCall("addEventListener",eventType, listener);
	
		var event = events[eventType];
		if (!eventType)
			error("addEventListener", "No event was given.");
		else if (!event)
			error("addEventListener", "An unrecognized event type was given ("+eventType+").");
		else if (!listener)
			error("addEventListener", "No listener was given.");
		else
			event.addListener(listener);
	};

	mraid.removeEventListener = function(eventType, listener) {
		traceCall("removeEventListener",eventType, listener);
		
		var event = events[eventType];
		if (!eventType)
			error("removeEventListener", "No event was given.");
		else if (!event)
			error("removeEventListener", "An unrecognized event type was given ("+eventType+").");
		else
			event.removeListener(listener);
	};
}());
